from ..utils import getEndpointUrl
from abstract_apis import asyncPostRequest,asyncGetRequest

async def async_get_rate_limit_url(method_name, *args, **kwargs):
    url = getEndpointUrl("rate_limit")
    kwargs["method"]=method_name or kwargs.get("method")
    for arg in args:
        if "method" not in kwargs:
            kwargs["method"] = arg
    return await asyncGetRequest(url, kwargs)



async def async_log_response(method_name, response_data,endpoint=None, *args, **kwargs):
    url = getEndpointUrl("/log_response")
    # Assuming that `args` is supposed to update `kwargs` for unspecified keys
    for arg in args:
        if "method" not in kwargs:
            kwargs["method"] = arg
        if "response_data" not in kwargs:
            kwargs["response_data"] = response_data
    # Use a direct payload if kwargs are not being used effectively
    kwargs["response_data"] = response_data
    kwargs["method"] = method_name

    return await asyncPostRequest(url, kwargs,endpoint=None)






from ..async_utils.abstract_async_rate_limit import async_log_response,async_get_rate_limit_url
from ..utils import getEndpointUrl
import asyncio
import asyncio

def get_async_response(func, *args, **kwargs):
    if asyncio.get_event_loop().is_running():
        # If already inside an event loop, use ensure_future or create_task
        return asyncio.ensure_future(func(*args, **kwargs))
    else:
        return asyncio.run(func(*args, **kwargs))

def get_rate_limit_url(method_name, *args, **kwargs):
    return get_async_response(async_get_rate_limit_url, method_name, *args, **kwargs)

def log_response(method_name, response_data, endpoint=None, *args, **kwargs):
    return get_async_response(async_log_response, method_name, response_data, endpoint=endpoint, *args, **kwargs)




from .abstract_transactions import *
from .address_lookup import *
from .abstract_swap import *
from .classes import *
from .abstract_logs import *
from .price_funcs import *
from .raydium_pool_keys import *
from .pumpFunKeys import *



from solana.rpc.async_api import AsyncClient
from solana.transaction import Transaction
from ..async_utils.abstract_async_rate_limit import *
from typing import List, Dict, Any
import asyncio
from .utils import *
def process_binary_data(binary_data,luts=[]):
    decoded_data = binary_data.data  # Accessing the raw data directly
    addresses = decode_lut_data(decoded_data)  # Use the implemented function here
    luts.extend([get_pubString(address) for address in addresses])
    return luts

def decode_lut_data(data: bytes) -> List[str]:
    """
    Decodes LUT (Address Lookup Table) data from the account's raw binary data.
    
    Args:
        data (bytes): The binary data retrieved from the LUT account.
    
    Returns:
        List[str]: A list of base58 encoded addresses.
    """
    # Assuming the data is base64 encoded
    address_size = 32  # Public keys in Solana are 32 bytes
    addresses = []

    # Process the binary data directly
    for i in range(0, len(data), address_size):
        address_bytes = data[i:i + address_size]
        if len(address_bytes) == address_size:
            address = get_pubBytes(address_bytes)
            addresses.append(address)

    return addresses

async def fetch_lut_addresses(client: AsyncClient, lut_keys: List[Pubkey]) -> List[Pubkey]:
    luts = []
    for lut_key in lut_keys:
        response = await client.get_account_info(lut_key)
        
        account_info = response.value
        if account_info:
            process_binary_data(account_info,luts)
    return luts

async def process_transaction(client: AsyncClient, txnData: Dict[str, Any]) -> List[str]:
    account_keys = get_all_account_keys(txnData)
    account_keys = [get_pubString(key) for key in account_keys]

    # Fetch and combine addresses from LUTs
    lut_keys = [get_pubString(acc['accountKey']) for acc in get_addres_lookup_table(txnData)]
    lut_addresses = await fetch_lut_addresses(client, lut_keys)

    # Combine the original account keys with LUT addresses
    all_accs = account_keys

    # Convert to base58 for consistency
    all_accs_base58 = [str(acc) for acc in all_accs]

    # Validate and sort account keys
    sorted_addresses = sorted(all_accs_base58)

    # Process each account key dynamically
    for address in sorted_addresses:
        account_index = all_accs_base58.index(address)

        # Example of fetching associated data (assuming you have a method to get it)
        account_data = get_post_balance_fromm_txn(txnData)
        relevant_data = next((item for item in account_data if item['accountIndex'] == account_index), None)

        #if relevant_data:
        #    print(f"Data for {address}: {relevant_data}")
        #else:
            #print(f"No data found for address: {address}")
    return sorted_addresses

async def main_sorted_addresses(txn):
    method = 'sorted_addresses'
    url = await get_rate_limit_url(method)
    client = AsyncClient(url.get('url'))
    sorted_addresses = await process_transaction(client, txn)
    await client.close()
    await log_response(method, sorted_addresses)
    return sorted_addresses  # Return the sorted addresses

# Modify the main function to use asyncio's event loop directly
def get_sorted_addresses(txnData):
    return asyncio.run(main_sorted_addresses(txnData))



from ..utils import *
def get_balance_info(balanceInfo, programId, authority):
    return next((bal for bal in balanceInfo if programId == bal.get("programId") and authority == bal.get("owner")), {})

def get_balance_from_instruction(transfer_instruction, txnData):
    preBalance = txnData["meta"].get("preTokenBalances", [])
    postBalance = txnData["meta"].get("postTokenBalances", [])
    
    # Update transfer_instruction with source, destination, and authority
    accounts = transfer_instruction.get("accounts", [])
    transfer_instruction.update({
        key: accounts[i] for i, key in enumerate(["source", "destination", "authority"])
    })
    
    preBalanceInfo = get_balance_info(preBalance, transfer_instruction['programId'], transfer_instruction['authority'])
    postBalanceInfo = get_balance_info(postBalance, transfer_instruction['programId'], transfer_instruction['authority'])
    
    amount = preBalanceInfo.get('uiTokenAmount', {}).get('amount', 0) - postBalanceInfo.get('uiTokenAmount', {}).get('amount', 0)
    transfer_instruction['amount'] = amount
    
    return transfer_instruction

def get_token_balances(txnData, pre=True, token=True):
    token_prefix = 'Token' if token else ''
    balance_type = 'pre' if pre else 'post'
    return txnData["meta"].get(f"{balance_type}{token_prefix}Balances", [])

def get_transfer_info_from_ids(txnData, address, token=True):
    tokenBalances = {"pre": [], "post": []}
    
    for pre in [True, False]:
        balanceInfos = get_token_balances(txnData, pre=pre, token=token)
        balanceKey = 'pre' if pre else 'post'
        tokenBalances[balanceKey].extend(
            txn for txn in balanceInfos if address in txn.values()
        )
    
    balances = []
    for tokenBalance in tokenBalances['post']:
        postBalance = tokenBalance.get('uiTokenAmount', {}).get('amount', 0)
        match = get_balance_info(tokenBalances['pre'], tokenBalance.get("programId"), tokenBalance.get("owner"))
        change = postBalance - match.get("uiTokenAmount", {}).get("amount", 0)
        tokenBalance['change'] = change
        balances.append(tokenBalance)
    
    return balances

def get_transfer_info(transfer_instructions, txnData):
    for transfer_instruction in transfer_instructions:
        updated_instruction = get_balance_from_instruction(transfer_instruction, txnData)
        transfer_instruction.update(updated_instruction)
    
    return transfer_instructions



from typing import List, Dict, Any, Optional, Union
from solders.pubkey import Pubkey
import base64
from ..utils import ifListGetSection
def get_pubString(obj):
    return Pubkey.from_string(str(obj))

def get_pubBytes(obj):
    return Pubkey.from_bytes(obj)

def get_addres_lookup_table(txnData):
    return txnData.get('transaction', {}).get('message', {}).get('addressTableLookups', [])

def get_account_keys(txnData):
    return txnData.get('transaction', {}).get('message', {}).get('accountKeys', [])

def get_loaded_addresses(txnData):
    return txnData['meta']['loadedAddresses']

def get_read_only_addresses(txnData):
    return get_loaded_addresses(txnData).get('readonly', [])

def get_writable_addresses(txnData):
    return get_loaded_addresses(txnData).get('writable', [])

def get_log_messages(txnData):
    return txnData['meta']['logMessages']

def get_instructions(txnData):
  return txnData['transaction']['message']['instructions']

def get_inner_instructions(txnData):
  return txnData['meta']['innerInstructions'][0]['instructions']

def update_instructions(txnData,instructions):
  txnData['transaction']['message']['instructions'] = instructions
  return  txnData

def update_inner_instructions(txnData,inner_instructions):
  txnData['meta']['innerInstructions'][0]['instructions'] = inner_instructions
  return txnData

def get_post_token_balances_from_txn(txnData):
    return txnData.get('meta', {}).get('postTokenBalances', [])

def get_pre_token_balances_from_txn(txnData):
    return txnData.get('meta', {}).get('preTokenBalances', [])

def get_post_balances_from_txn(txnData):
    return txnData.get('meta', {}).get('postBalances', [])

def get_pre_balances_from_txn(txnData):
    return txnData.get('meta', {}).get('preBalances', [])

def get_all_account_keys(txnData):
  accountKeys=[]
  accountKeys += get_account_keys(txnData)
  accountKeys += get_read_only_addresses(txnData)
  accountKeys += get_writable_addresses(txnData)
  return accountKeys

def search_for_account_index(data,index_number):
    for index_data in data:
        if str(index_data.get('accountIndex')) == str(index_number):
            return index_data
        
def extract_lp_initialization_log_entry_info_from_log_entry(lp_log_entry: str) -> Dict[str, Union[int, float]]:
    lp_initialization_log_entry_info_start = lp_log_entry.find('{')
    return json.loads(fix_relaxed_json_in_lp_log_entry(lp_log_entry[lp_initialization_log_entry_info_start:]))

def fix_relaxed_json_in_lp_log_entry(relaxed_json: str) -> str:
    return relaxed_json.replace(r'([{,])\s*([a-zA-Z_][a-zA-Z0-9_]*)\s*:', r'\1"\2":')

def find_log_entry(needle: str, log_entries: List[str]) -> Optional[str]:
    for log_entry in log_entries:
        if needle in log_entry:
            return log_entry
    return None



from abstract_utilities import make_list
from ..utils import *
class logMessageManager:
      def __init__(self,txnData):
        self.logMessages = get_log_messages(txnData)
        self.logClusters = self.get_log_index_clusters()
      def get_log_index_clusters(self):
        clusters =[]
        clusterCount=-1
        for j,log in enumerate(self.logMessages):
            curr_logs = {}
            
            # Split the log into parts based on whitespace
            parts = [part.replace(':','') for part in log.split()]
            
            if not parts:
                continue  # Skip empty lines
            # Parse based on the structure
            if "invoke" in parts:
                
                clusterCount+=1
                clusters.append({"logs":[],"clusterIndex":clusterCount,"programId":parts[1],"stackHeight":str(parts[-1])[1:-1],"start":j,"end":j,"types":[]})
                
                curr_logs['cmdType'] = parts[0]
                curr_logs['programId'] = parts[1]
                curr_logs['event'] = parts[2]  # 'invoke' or 'success'
                curr_logs['vars'] = parts[3] if len(parts) > 3 else None
            
            elif "consumed" in parts:
                curr_logs['cmdType'] = parts[0]
                curr_logs['programId'] = parts[1]
                curr_logs['event'] = 'consumed'
                consumed_index = parts.index('consumed')
                curr_logs['vars'] = " ".join(parts[consumed_index + 1:])
            
            elif "return:" in log:
                curr_logs['cmdType'] = parts[0]
                curr_logs['programId'] = parts[1]
                curr_logs['event'] = 'return'
                curr_logs['vars'] = " ".join(parts[2:])
            
            elif "log:" in log:
                curr_logs['cmdType'] = parts[0]
                curr_logs['programId'] = parts[1]
                log_index = parts.index('log')
                curr_logs['event'] = parts[log_index + 1]
                curr_logs['type'] = parts[log_index + 2] if len(parts) > log_index + 2 else None
                clusters[-1]['types'].append(curr_logs['type'])
                if clusters[-1].get('type') == None:
                  clusters[-1]['type']=curr_logs['type']
                curr_logs['vars'] = " ".join(parts[log_index + 3:]) if len(parts) > log_index + 3 else None
            
            else:
                curr_logs['cmdType'] = parts[0]
                curr_logs['programId'] = parts[1]
                curr_logs['event'] = parts[2] if len(parts) > 2 else None
                curr_logs['vars'] = " ".join(parts[3:]) if len(parts) > 3 else None
            clusters[-1]['end']=j
            # Store parsed log in the list
            clusters[-1]["logs"].append(curr_logs)
        return clusters      
      def get_cluster_type_from_index(self,index):
        cluster = {}
        if len(self.logClusters) >index:
          cluster = self.logClusters[index]
        clusterType = make_list(cluster.get("type") or cluster.get('types'))
        if clusterType:
          clusterType = clusterType[0]
        return clusterType
      def add_cluster_type_to_instruction(self,index,instruction):
        clusterType = self.get_cluster_type_from_index(index)
        instruction['type']=clusterType
        return instruction
      
      def add_cluster_types_to_instructions(self,instructions,indexStart=0):
        new_instructions = []
        for i,instruction in enumerate(instructions):
          index = i+indexStart
          instruction = self.add_cluster_type_to_instruction(index,instruction)
          new_instructions.append(instruction)
        return new_instructions



from .. import *
def get_instruction_map():
  return [{'instruction_number': '1', 'instruction_name': 'Mint', 'token_address': 'AkAUSJg1v9xYT3HUxdALH7NsrC6owmwoZuP9MLw8fxTL', 'token_name': '3CAT'}, {'instruction_number': '2', 'instruction_name': 'Mint Authority', 'token_address': 'TSLvdd1pWpHVjahSpsvCXUbgwsL3JAcvokwaKt1eokM', 'token_name': 'Pump.fun Token Mint Authority'}, {'instruction_number': '3', 'instruction_name': 'Bonding Curve', 'token_address': '9nhxvNxfSUaJddVco6oa6NodtsCscqCScp6UU1hZkfGm', 'token_name': 'Pump.fun (3CAT) Bonding Curve'}, {'instruction_number': '4', 'instruction_name': 'Associated Bonding Curve', 'token_address': '889XLp3qvVAHpTYQmhn6cBpYSppV8Gi8E2Rgp9RH2vRy', 'token_name': 'Pump.fun (3CAT) Vault'}, {'instruction_number': '5', 'instruction_name': 'Global', 'token_address': '4wTV1YmiEkRvAtNtsSGPtUrqRYQMe5SKy2uB4Jjaxnjf', 'token_name': '4wTV1YmiEkRvAtNtsSGPtUrqRYQMe5SKy2uB4Jjaxnjf'}, {'instruction_number': '6', 'instruction_name': 'Mpl Token Metadata', 'token_address': 'metaqbxxUerdq28cj1RbAWkYQm3ybzjb6a8bt518x1s', 'token_name': 'Metaplex Token Metadata'}, {'instruction_number': '7', 'instruction_name': 'Metadata', 'token_address': 'CH41RxpjSXHqr1vfLTVYJMsfNs2fBCCWoAE13tPihXh7', 'token_name': 'CH41RxpjSXHqr1vfLTVYJMsfNs2fBCCWoAE13tPihXh7'}, {'instruction_number': '8', 'instruction_name': 'User', 'token_address': 'Fuy5MvbgzjSok1U8hH6mUY6WnLynzUextDxfEWMiTkn4', 'token_name': 'Fuy5MvbgzjSok1U8hH6mUY6WnLynzUextDxfEWMiTkn4'}, {'instruction_number': '9', 'instruction_name': 'System Program', 'token_address': '11111111111111111111111111111111', 'token_name': 'System Program'}, {'instruction_number': '10', 'instruction_name': 'Token Program', 'token_address': 'TokenkegQfeZyiNwAJbNbGKPFXCWuBvf9Ss623VQ5DA', 'token_name': 'Token Program'}, {'instruction_number': '11', 'instruction_name': 'Associated Token Program', 'token_address': 'ATokenGPvbdGVxr1b2hvZbsiqW5xWH25efTNsLJA8knL', 'token_name': 'Associated Token Account Program'}, {'instruction_number': '12', 'instruction_name': 'Rent', 'token_address': 'SysvarRent111111111111111111111111111111111', 'token_name': 'Rent Program'}, {'instruction_number': '13', 'instruction_name': 'Event Authority', 'token_address': 'Ce6TQqeHC9p8KetsN6JsjHK7UTZk7nasjjnr7XxXp9F1', 'token_name': 'Ce6TQqeHC9p8KetsN6JsjHK7UTZk7nasjjnr7XxXp9F1'}, {'instruction_number': '14', 'instruction_name': 'Program', 'token_address': '6EF8rrecthR5Dkzon8Nwu78hRvfCKubJ14M5uBEwF6P', 'token_name': 'Pump.fun'}]
def get_pump_signature():
  return "dpL54KTyCPJQmm5cQmKm8nb38r9y7CGthR6D1CczTLMT5UvbxXwVy7FRzrCuHncJmGY6PqEjgiF8MTZg1YX1b7X"
def get_signature_js(signature=None):
  return {"signature":signature or get_pump_signature()}

def build_instructions_template(txnData,instruction_maps=None,printIt=False,saveIt=False):
  instruction_maps = instruction_maps or get_instruction_map()
  instructions = get_instructions_catalog(txnData,printIt=printIt,saveIt=saveIt)
  new_token_js = {}

  for j,instruction_map in enumerate(instruction_maps):
    addressType = instruction_map.get('instruction_name')
    address = instruction_map.get('token_address')
    instruction_maps[j]["catalog"]=[]
    found = False
    for instruction in instructions:
      inst_accounts = instruction.get('accounts')
      if address in inst_accounts:
        for i,inst_address in enumerate(inst_accounts):
          if inst_address == address:
            if found:
              if printIt:
                print(f"another catalog found for {address}")
            else:
              if printIt:
                print(f"#{j} - {addressType} FOUND")
            found = True
            if printIt:
              print(f"{addressType} found")
              print(f"event {instruction.get('events')}")
              print(f"accountKey {i}")
              print(f"programId {instruction.get('programId')}\n")
            instruction_maps[j]["catalog"].append({"index":i,"catalog":instruction})
    if found == False:
      print(f"#{addressType} NOT FOUND")
    j+=1
  return instruction_maps
def getPumpMap(signature = get_pump_signature()):
  txnData = getTransaction(signature=signature)
  instruction_maps = build_instructions_template(txnData,instruction_maps=None,printIt=False,saveIt=False)
  new_maps = []
  for instruction_map in instruction_maps:
    address = instruction_map.get('token_address')
    name = instruction_map.get('token_name')
    instruction_name = instruction_map.get('instruction_name')
    instruction_index = instruction_map.get('instruction_number')
    getEvent={}
    for catalogs in instruction_map.get('catalog'):
      account_index= catalogs.get('index')
      catalog = catalogs.get('catalog')
      events = catalog.get('events')
      accounts = catalog.get('accounts')
      if events:
        if accounts[account_index] == address:
          getEvent = {"event":events[0],"account":account_index,"address_type":instruction_name}
          if events[0] != 'Create':
            break
    new_maps.append(getEvent)
  return new_maps
def getTxnTypes(signature,maps=getPumpMap()):
  token_types = {}
  txnData = getTransaction(signature)
  instructions = get_instructions_catalog(txnData,printIt=False,saveIt=False)
  for new_map in maps:
    for instruction in instructions:
      if new_map.get('event') in instruction.get('events'):
        account = instruction.get('accounts')[new_map.get('account')]
        token_types[new_map.get('address_type')]=account
        break
  return token_types
